/* $Id: default.c,v 3.16 1993/06/28 20:53:33 bjoerns Exp $
 *
 *	This file is part of the XPilot project, written by
 *
 *	    Bjrn Stabell (bjoerns@staff.cs.uit.no)
 *	    Ken Ronny Schouten (kenrsc@stud.cs.uit.no)
 *	    Bert Gsbers (bert@mc.bio.uva.nl)
 *
 *	Copylefts are explained in the LICENSE file.
 */

#include <X11/Xos.h>
#include <X11/keysym.h>
#ifdef	apollo
#    include <X11/ap_keysym.h>
#endif
#include "config.h"
#include "const.h"
#include "object.h"
#include "client.h"
#include "paint.h"
#include "pack.h"
#include "bit.h"
#include "netclient.h"

#ifndef	lint
static char sourceid[] =
    "@(#)$Id: default.c,v 3.16 1993/06/28 20:53:33 bjoerns Exp $";
#endif

/*
 * Default fonts
 */
#define GAME_FONT		"-*-times-*-*-*-*-18-*-*-*-*-*-iso8859-1"
#define MESSAGE_FONT		"-*-times-*-*-*-*-14-*-*-*-*-*-iso8859-1"
#define SCORE_LIST_FONT		"-*-fixed-bold-*-*-*-15-*-*-*-c-*-iso8859-1"
#define BUTTON_FONT		"-*-*-bold-o-*-*-14-*-*-*-*-*-iso8859-1"
#define TEXT_FONT		"-*-*-bold-i-*-*-14-*-*-*-p-*-iso8859-1"


static struct _keyResources
{
    char    *resource;
    keys_t	key;
    char	*helpLine;
} keyResources[] = {
    "keyTurnLeft",	KEY_TURN_LEFT,		"Turn left (anti-clockwise)",
    "keyTurnRight",	KEY_TURN_RIGHT,		"Turn right (clockwise)",
    "keyThrust",	KEY_THRUST,		"Thrust",
    "keyShield",	KEY_SHIELD,		"Raise shield",
    "keyFireShot",	KEY_FIRE_SHOT,		"Fire shot",
    "keyFireMissile",	KEY_FIRE_MISSILE,	"Fire smart missile",
    "keyFireTorpedo",	KEY_FIRE_TORPEDO,	"Fire unguided torpedo",
    "keyFireHeat",	KEY_FIRE_HEAT,		"Fire heat seaking missile",
    "keyFireNuke",	KEY_FIRE_NUKE,		"Fire nuclear missile",
    "keyDropMine",	KEY_DROP_MINE,		"Drop a mine (bomb)",
    "keyDetachMine",	KEY_DETACH_MINE,	"Detach a mine",
    "keyLockClose",	KEY_LOCK_CLOSE,		"Lock on closest player",
    "keyLockNext",	KEY_LOCK_NEXT,		"Lock on next player",
    "keyLockPrev",	KEY_LOCK_PREV,		"Lock on previous player",
    "keyRefuel",	KEY_REFUEL,		"Refuel",
    "keyCloak",		KEY_CLOAK,		"Toggle cloakdevice",
    "keyEcm",		KEY_ECM,		"Use ECM",
    "keySelfDestruct",	KEY_SELF_DESTRUCT,	"Toggle self destruct",
    "keyIdMode",	KEY_ID_MODE,		"Toggle ID mode",
    "keyPause",		KEY_PAUSE,		"Toggle pause mode",
    "keySwapSettings",	KEY_SWAP_SETTINGS,	"Swap control settings",
    "keyChangeHome",	KEY_CHANGE_HOME,	"Change home base",
    "keyConnector",	KEY_CONNECTOR,		"Use connector (pick up ball)",
    "keyDropBall",	KEY_DROP_BALL,		"Drop a ball",
    "keyTankNext",	KEY_TANK_NEXT,		"Shift to next tank",
    "keyTankPrev",	KEY_TANK_PREV,		"Shift to previous tank",
    "keyTankDetach",	KEY_TANK_DETACH,	"Detach tank",
    "keyIncreasePower",	KEY_INCREASE_POWER,	"Increase power",
    "keyDecreasePower",	KEY_DECREASE_POWER,	"Decrease power",
    "keyIncreaseTurnspeed",KEY_INCREASE_TURNSPEED,	"Increase turnspeed",
    "keyDecreaseTurnspeed",KEY_DECREASE_TURNSPEED,	"Decrease turnspeed",
    "keyTransporter",	KEY_TRANSPORTER,	"Use transporter",
    "keyToggleVelocity",KEY_TOGGLE_VELOCITY,	"N/A",
    "keyToggleCompass",	KEY_TOGGLE_COMPASS,	"N/A",
};


static int ON(char *optval)
{
    return (strcasecmp(optval, "true") == 0
	    || strcasecmp(optval, "on") == 0
	    || strcasecmp(optval, "yes") == 0);
}


static float atod(char *str)
{
    float tmp = 0.0;

    sscanf(str, "%f", &tmp);
    return (tmp);
}


char* Get_keyhelpstring(keys_t key)
{
    int i=0;
    char* str = NULL;

    for (i = 0; i < NELEM(keyResources); i++)
	if (keyResources[i].key == key) {
	    str = keyResources[i].helpLine;
	    break;
	}

    return str;
}


void Get_defaults(void)
{
    char		*str;
    int			i;
    static keydefs_t defaultKeyDefs[] = {
	XK_a,			KEY_TURN_LEFT,
	XK_s,			KEY_TURN_RIGHT,
	XK_Shift_R,		KEY_THRUST,
	XK_space,		KEY_SHIELD,
	XK_Return,		KEY_FIRE_SHOT,
	XK_backslash,		KEY_FIRE_MISSILE,
	XK_quoteright,		KEY_FIRE_TORPEDO,
	XK_semicolon,		KEY_FIRE_HEAT,
	XK_n,			KEY_FIRE_NUKE,
	XK_Tab,			KEY_DROP_MINE,
	XK_bracketright,	KEY_DETACH_MINE,
	XK_Select,		KEY_LOCK_CLOSE,
	XK_Next,		KEY_LOCK_NEXT,
	XK_Prior,		KEY_LOCK_PREV,
	XK_Home,		KEY_CHANGE_HOME,
	XK_Control_L,		KEY_REFUEL,
	XK_bracketleft,		KEY_ECM,
	XK_Delete,		KEY_CLOAK,
	XK_t,			KEY_TRANSPORTER,
	XK_Control_L,		KEY_CONNECTOR,
	XK_d,			KEY_DROP_BALL,
	XK_q,			KEY_SELF_DESTRUCT,
	XK_i,			KEY_ID_MODE,
	XK_p,			KEY_PAUSE,
	XK_v,			KEY_TOGGLE_VELOCITY,
	XK_c,			KEY_TOGGLE_COMPASS,
	XK_Escape,		KEY_SWAP_SETTINGS,
	XK_KP_Multiply,		KEY_INCREASE_POWER,
	XK_KP_Divide,		KEY_DECREASE_POWER,
	XK_KP_Add,		KEY_INCREASE_TURNSPEED,
	XK_KP_Subtract,		KEY_DECREASE_TURNSPEED,
	XK_w,			KEY_TANK_PREV,
	XK_e,			KEY_TANK_NEXT,
	XK_r,			KEY_TANK_DETACH,
    };


    /*
     * Defaults.
     */
    power		= 45.0;
    turnspeed		= 30.0;
    turnresistance	= 0.12;

    power_s		= 35.0;
    turnspeed_s		= 25.0;
    turnresistance_s	= 0.12;
    fuelLevel3		= 500;
    fuelLevel2		= 200;
    fuelLevel1		= 100;
    instruments		= SHOW_HUD_INSTRUMENTS | SHOW_HUD_HORIZONTAL;
    instruments		|= SHOW_SHIP_NAME;
    memcpy(keyDefs, defaultKeyDefs, sizeof(defaultKeyDefs));
    receive_window_size	= DEF_RECEIVE_WINDOW_SIZE;

    /*
     * Name
     */
    if (name[0] == '\0') {
	if ((dpy != NULL)
	    && (str = XGetDefault(dpy, "xpilot", "name")) != NULL) {
	    strncpy(name, str, MAX_NAME_LEN);
	    name[MAX_NAME_LEN-1] = '\0';
	} else
	    strcpy(name, realname);
    }
    CAP_LETTER(name[0]);

#ifdef SOUND
    str = XGetDefault(dpy, "xpilot", "sounds");
    sounds = str ? str : SOUNDFILE;

    str = XGetDefault(dpy, "xpilot", "maxVolume");
    maxVolume = str ? atoi(str) : 100;
#endif /* SOUND */

    /*
     * Team
     */
    if (team == TEAM_NOT_SET) {
	str = XGetDefault(dpy, "xpilot", "team");
	if (str)
	    team = atoi(str);
    }

    /*
     * Control sensitivity
     */
    str = XGetDefault(dpy, "xpilot", "power");
    if (str)
	power = atod(str);
    str = XGetDefault(dpy, "xpilot", "turnSpeed");
    if (str)
	turnspeed = atod(str);
    str = XGetDefault(dpy, "xpilot", "turnResistance");
    if (str)
	turnresistance = atod(str);

    str = XGetDefault(dpy, "xpilot", "altPower");
    if (str)
	power_s = atod(str);
    str = XGetDefault(dpy, "xpilot", "altTurnSpeed");
    if (str)
	turnspeed_s = atod(str);
    str = XGetDefault(dpy, "xpilot", "altTurnResistance");
    if (str)
	turnresistance_s = atod(str);

    /*
     * Visual appearance of objects on screen.
     */
    str = XGetDefault(dpy, "xpilot", "backgroundPointDist");
    if (str == NULL || (map_point_distance = atoi(str)) <= 0) {
	map_point_distance = 8;
    }

    /*
     * HUD (Head Up Display) related parameters
     */
    str = XGetDefault(dpy, "xpilot", "showHUD");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_HUD_INSTRUMENTS);
	} else {
	    CLR_BIT(instruments, SHOW_HUD_INSTRUMENTS);
	}
    }
    str = XGetDefault(dpy, "xpilot", "verticalHUDLine");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_HUD_VERTICAL);
	} else {
	    CLR_BIT(instruments, SHOW_HUD_VERTICAL);
	}
    }
    str = XGetDefault(dpy, "xpilot", "horizontalHUDLine");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_HUD_HORIZONTAL);
	} else {
	    CLR_BIT(instruments, SHOW_HUD_HORIZONTAL);
	}
    }
    str = XGetDefault(dpy, "xpilot", "speedFactHUD");
    hud_move_fact = str?atod(str):0.0;
    str = XGetDefault(dpy, "xpilot", "speedFactPTR");
    ptr_move_fact = str?atod(str):0.0;


    /*
     * Fuel related parameters
     */
    str = XGetDefault(dpy, "xpilot", "fuelMeter");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_FUEL_METER);
	} else {
	    CLR_BIT(instruments, SHOW_FUEL_METER);
	}
    }
    str = XGetDefault(dpy, "xpilot", "fuelGauge");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_FUEL_GAUGE);
	} else {
	    CLR_BIT(instruments, SHOW_FUEL_GAUGE);
	}
    }
    str = XGetDefault(dpy, "xpilot", "fuelNotify");
    if (str)
	fuelLevel3 = atod(str);
    str = XGetDefault(dpy, "xpilot", "fuelWarning");
    if (str)
	fuelLevel2 = atod(str);
    str = XGetDefault(dpy, "xpilot", "fuelCritical");
    if (str)
	fuelLevel1 = atod(str);

    /*
     * Misc. meters
     */
    str = XGetDefault(dpy, "xpilot", "turnSpeedMeter");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_TURNSPEED_METER);
	} else {
	    CLR_BIT(instruments, SHOW_TURNSPEED_METER);
	}
    }
    str = XGetDefault(dpy, "xpilot", "powerMeter");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_POWER_METER);
	} else {
	    CLR_BIT(instruments, SHOW_POWER_METER);
	}
    }
    str = XGetDefault(dpy, "xpilot", "packetSizeMeter");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_PACKET_SIZE_METER);
	} else {
	    CLR_BIT(instruments, SHOW_PACKET_SIZE_METER);
	}
    }
    str = XGetDefault(dpy, "xpilot", "packetLossMeter");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_PACKET_LOSS_METER);
	} else {
	    CLR_BIT(instruments, SHOW_PACKET_LOSS_METER);
	}
    }
    str = XGetDefault(dpy, "xpilot", "packetDropMeter");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_PACKET_DROP_METER);
	} else {
	    CLR_BIT(instruments, SHOW_PACKET_DROP_METER);
	}
    }

    /*
     * Radar related parameters
     */
    str = XGetDefault(dpy, "xpilot", "slidingRadar");
    if (str) {
	if (ON(str)) {
	    SET_BIT(instruments, SHOW_SLIDING_RADAR);
	} else {
	    CLR_BIT(instruments, SHOW_SLIDING_RADAR);
	}
    }

    /*
     * Fonts
     */
    if (str = XGetDefault(dpy, "xpilot", "gameFont"))
	strcpy(gameFontName, str);
    else
	strcpy(gameFontName, GAME_FONT);

    if (str = XGetDefault(dpy, "xpilot", "messageFont"))
	strcpy(messageFontName, str);
    else
	strcpy(messageFontName, MESSAGE_FONT);

    if (str = XGetDefault(dpy, "xpilot", "scoreListFont"))
	strcpy(scoreListFontName, str);
    else
	strcpy(scoreListFontName, SCORE_LIST_FONT);

    if (str = XGetDefault(dpy, "xpilot", "buttonFont"))
	strcpy(buttonFontName, str);
    else
	strcpy(buttonFontName, BUTTON_FONT);

    if (str = XGetDefault(dpy, "xpilot", "textFont"))
	strcpy(textFontName, str);
    else
	strcpy(textFontName, TEXT_FONT);

    /*
     * Network performance tuning.
     */
    str = XGetDefault(dpy, "xpilot", "receiveWindowSize");
    if (str) {
	if (sscanf(str, "%d", &receive_window_size) != 1
	    || receive_window_size < MIN_RECEIVE_WINDOW_SIZE
	    || receive_window_size > MAX_RECEIVE_WINDOW_SIZE) {
	    printf("%s is not a valid receive window size.\n", str);
	    receive_window_size = DEF_RECEIVE_WINDOW_SIZE;
	}
    }

    /*
     * Key bindings
     */
    for (i = 0; i < NELEM(keyResources); i++)
	if (str = XGetDefault(dpy, "xpilot", keyResources[i].resource)) {
	    KeySym ks;
	    int j;

	    if ((ks = XStringToKeysym(str)) == NoSymbol) {
		printf("%s is not a valid keysym.\n", str);
		continue;
	    }

	    for (j = 0; j < MAX_KEY_DEFS; j++)
		if (!keyDefs[j].keysym
		    || keyDefs[j].key == keyResources[i].key) {
		    keyDefs[j].keysym = ks;
		    keyDefs[j].key = keyResources[i].key;
		    break;
		}

	    if (j >= MAX_KEY_DEFS) {
		printf("Too may key definitions.  Only %d allowed.\n",
		       MAX_KEY_DEFS);
		i = NELEM(keyResources);
		break;
	    }
	}
}
