/* $Id: object.h,v 3.9 1993/06/28 20:53:57 bjoerns Exp $
 *
 *	This file is part of the XPilot project, written by
 *
 *	    Bjrn Stabell (bjoerns@staff.cs.uit.no)
 *	    Ken Ronny Schouten (kenrsc@stud.cs.uit.no)
 *	    Bert Gsbers (bert@mc.bio.uva.nl)
 *
 *	Copylefts are explained in the LICENSE file.
 */

#ifndef	OBJECT_H
#define	OBJECT_H

#include "const.h"
#include "types.h"
#include "bit.h"
#include "keys.h"


#define OBJ_PLAYER		(1UL<<0)	/* Types of objects */
#define OBJ_DEBRIS		(1UL<<1)
#define OBJ_SPARK		(1UL<<2)
#define OBJ_NUKE                (1L<<3)
#define OBJ_BALL		(1UL<<7)
#define OBJ_SHOT		(1UL<<8)	/* Misc. objects */
#define OBJ_SMART_SHOT		(1UL<<9)
#define OBJ_ROCKET_PACK		(1UL<<10)
#define OBJ_CLOAKING_DEVICE	(1UL<<11)
#define OBJ_ENERGY_PACK		(1UL<<12)
#define OBJ_WIDEANGLE_SHOT	(1UL<<13)
#define OBJ_TRAINER		(1UL<<14)
#define OBJ_SHIELD		(1UL<<15)
#define OBJ_REFUEL		(1UL<<16)
#define OBJ_COMPASS		(1UL<<17)
#define OBJ_BACK_SHOT		(1UL<<18)
#define OBJ_MINE		(1UL<<19)
#define OBJ_MINE_PACK		(1UL<<20)
#define OBJ_SENSOR_PACK		(1UL<<21)
#define OBJ_TANK		(1UL<<22)
#define OBJ_ECM			(1UL<<23)
#define OBJ_TORPEDO		(1UL<<24)
#define OBJ_HEAT_SHOT		(1UL<<25)
#define OBJ_AFTERBURNER	(1UL<<26)
#define OBJ_CONNECTOR		(1UL<<27)
#define OBJ_TRANSPORTER         (1UL<<28)
#define OBJ_FIRE		(1UL<<29)
#define OBJ_CANNON_DEBRIS	(1UL<<30)	/* Cannon objects */
#define OBJ_CANNON_SHOT		(1UL<<31)

/*
 * Possible player status bits
 */
#define PLAYING			(1L<<0)		/* Not returning to base */
#define WAITING_SHOTS		(1L<<1)
#define KILLED			(1L<<2)
#define THRUSTING		(1L<<3)
#define SELF_DESTRUCT		(1L<<4)
#define GRAVITY			(1L<<5)
#define SHOT_GRAVITY		(1L<<6)
#define ID_MODE			(1L<<8)
#define LOOSE_MASS		(1L<<9)
#define PAUSE			(1L<<10)
#define GAME_OVER		(1L<<11)
#define INACTIVE		(1L<<12)    /* Accept keyboard input etc... */
#define FUEL_GAUGE		(1L<<13)
#define VELOCITY_GAUGE		(1L<<14)
#define POWER_GAUGE		(1L<<15)
#define WARPING			(1L<<16)
#define WARPED			(1L<<17)
#define CONFUSED		(1L<<18)

#define LOCK_NONE		1
#define LOCK_PLAYER		2

#define NOT_CONNECTED		(-1)

typedef struct {
    byte	color;			/* Color of object */		
    int		id;			/* For shots => id of player */
    int		placed;			/* The object has just been placed */
    int		wrapped;		/* The object is on the screen edge */
    position	prevpos;		/* Object's previous position... */
    position	pos;			/* World coordinates */
    ipos	intpos;
    vector	vel;
    vector	acc;
    int		dir;
    float	max_speed;
    float	velocity;
    float	turnspeed;
    float	mass;
    int		type;
    long	info;			/* Miscellaneous info */
    int		life;			/* No of ticks left to live */
    int		count;			/* Misc timings */
    long	status;

    int 	owner;			/* Who's object is this ? */
                                        /* (spare for id)*/
    int		treasure;		/* Which treasure does ball belong */
    int		new_info;
    float	length;			/* Distance between ball and player */
} object;


/*
 * Fuel structure, used by player
 */
typedef struct {
    long	sum;			/* Sum of fuel in all tanks */
    long	max;			/* How much fuel can you take? */
    int		current;		/* Number of currently used tank */
    int		num_tanks;		/* Number of tanks */
    long	tank[MAX_TANKS];
    long	count;			/* Display fuel for how long? */
    long	l1;			/* Fuel critical level */
    long	l2;			/* Fuel warning level */
    long	l3;			/* Fuel notify level */
} pl_fuel_t;

/* IMPORTANT
 *
 * This is the player structure, the first part MUST be similar to object_t,
 * this makes it possible to use the same basic operations on both of them
 * (mainly used in update.c).
 */
typedef struct {
    byte	color;			/* Color of object */		
    int		id;			/* Unique id of object */
    int		placed;			/* the object has just been placed */
    int		wrapped;		/* object crossed the world's edge */
    position	prevpos;		/* Previous position... */
    position	pos;			/* World coordinates */
    ipos	intpos;
    vector	vel;			/* Velocity of object */
    vector	acc;			/* Acceleration constant */
    int		dir;			/* Direction of acceleration */
    float	max_speed;		/* Maximum speed of object */
    float	velocity;		/* Absolute speed */
    float	turnspeed;		/* How fast player acc-turns */
    float	mass;			/* Mass of object (incl. cargo) */
    int		type;			/* Type of object */
    long	info;			/* Miscellaneous info */
    int		life;			/* Zero is dead. One is alive */
    int		count;			/* Miscellaneous timings */
    long	status;			/** Status, currently **/

    long	used;			/** Items you use **/
    long	have;			/** Items you have **/

    int		shield_time;		/* Shields if no playerShielding */
    pl_fuel_t	fuel;			/* ship tanks and the stored fuel */
    int		afterburners;		/* number of after burners, powerfull*/
					/* and efficient engine		     */
    float	emptymass;		/* Mass of empty ship */
    float	float_dir;		/* Direction, in float var */
    float	turnresistance;		/* How much is lost in % */
    float	turnvel;		/* Current velocity of turn (right) */
#ifdef TURN_FUEL
    float	oldturnvel;		/* Last velocity of turn (right) */
#endif
    float	turnacc;		/* Current acceleration of turn */
    long	mode;			/* Player mode, currently */
    long	instruments;		/* Instruments on screen (bitmask) */
    long	score;			/* Current score of player */
    long	prev_score;		/* Last score that has been updated */
    int		prev_life;		/* Last life that has been updated */
    float	power;			/* Force of thrust */
    float	power_s;		/* Saved power fiks */
    float	turnspeed_s;		/* Saved turnspeed */
    float	hud_move_fact;		/* scale the hud-movement (speed) */
    float	ptr_move_fact;		/* scale thes peed pointer length */
    float	turnresistance_s;	/* Saved (see above) */
    float	sensor_range;		/* Range of sensors (radar) */
    int		shots;			/* Number of active shots by player */
    int		extra_shots;		/* Number of extra shots / 2 */
    int		back_shots;		/* Number of rear shots */
    int		mines;			/* Number of mines. */
    int		cloaks;			/* Number of cloaks. */
    int		sensors;		/* Number of sensors */
    int		missiles;		/* Number of missiles. */
    int		ecms;			/* Number of ecms. */
    int		transporters;		/* number of transporters */
    int		shot_max;		/* Maximum number of shots active */
    int		shot_life;		/* Number of ticks shot will live */
    float	shot_speed;		/* Speed of shots fired by player */
    float	shot_mass;		/* Mass of shots fired by player */
    long	shot_time;		/* Time of last shot fired by player */
    int		fs;			/* Connected to fuel station fs */
    int		check;			/* Next check point to pass */
    int		time;			/* The time aa player has used */
    int		round;			/* Number of rounds player have done */
    int		best_lap;		/* Players best lap time */
    int		best_run;		/* Best race time */
    int		last_lap;		/* Time on last pass */
    int		last_lap_time;		/* What was your last pass? */
    int		last_time;		/* What was the time? */
    position	world;			/* Lower left hand corner is this */
					/* world coordinate */
    position	realWorld;		/* If the player is on the edge of
					   the screen, these are the world
					   coordinates before adjustment... */
    int		wrappedWorld;		/* Nonzero if we're near the edge
					   of the world (realWorld != world) */

    int		home_base;		/* Num of home base */
    struct {
	int	    tagged;		/* Flag, what is tagged? */
	int	    pl_id;		/* Tagging player id */
	position    pos;		/* Position of locked object */
	float	    distance;		/* Distance to object */
    } lock;

    char	mychar;			/* Special char for player */
    char	name[MAX_CHARS];	/* Nick-name of player */
    char	realname[MAX_CHARS];	/* Real name of player */
    char	dispname[MAX_CHARS];	/* Display name of player */
    int		map_point_distance;	/* spacing of navigation points */
    u_short	team;			/* What team is the player on? */
    u_short	pseudo_team;		/* Which team is used for my tanks */
					/* (detaching!) */
    object	*ball;
    /*
     * Robot variables
     */
    u_byte	robot_mode;		/* For players->RM_NOT_ROBOT */
    long	robot_count;		/* Misc timings, minimizes rand()use */
    int		robot_ind;		/* Index in the robot array */
    int		robot_lock;
    int		robot_lock_id;

    struct _visibility {
	int	canSee;
	long	lastChange;
    } *visibility;

    int updateVisibility, forceVisible, damaged;
    int wormDrawCount, wormHoleHit, wormHoleDest;

    struct {
	int size;
	position pos;
    } ecmInfo;

    struct {
	int count, pl_id;
    } transInfo;

    int	rplay_fd;			/* rplay UDP socket fd */

    int conn;				/* connection index, -1 if robot */

    BITV_DECL(last_keyv, NUM_KEYS);	/* Keyboard state */
    BITV_DECL(prev_keyv, NUM_KEYS);	/* Keyboard state */
    int key_changed;

    void *audio;			/* audio private data */
} player;

#endif
