/* $Id: server.c,v 3.15 1993/06/28 20:54:20 bjoerns Exp $
 *
 *	This file is part of the XPilot project, written by
 *
 *	    Bjrn Stabell (bjoerns@staff.cs.uit.no)
 *	    Ken Ronny Schouten (kenrsc@stud.cs.uit.no)
 *	    Bert Gsbers (bert@mc.bio.uva.nl)
 *
 *	Copylefts are explained in the LICENSE file.
 */

#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>
#ifndef __hpux
#include <sys/time.h>
#endif
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "global.h"
#include "socklib.h"
#include "version.h"
#include "map.h"
#include "pack.h"
#include "draw.h"
#include "robot.h"
#include "saudio.h"
#include "bit.h"
#include "netserver.h"

#ifndef	lint
static char versionid[] = "@(#)$" TITLE " $";
static char sourceid[] =
    "@(#)$Id: server.c,v 3.15 1993/06/28 20:54:20 bjoerns Exp $";
#endif


/*
 * Global variables
 */
int		NumPlayers = 0;
int		NumPseudoPlayers = 0;
int		NumObjs = 0;
int		Num_alive = 0;
player		**Players;
object		*Obj[MAX_TOTAL_SHOTS];
long		Id = 1;		    /* Unique ID for each object */
long		GetInd[MAX_ID];
server		Server;
int		Shutdown = -1, ShutdownDelay = 1000;
jmp_buf		SavedEnv;
int 		framesPerSecond = 18;

int	Argc;
char	**Argv;

static int	Socket;
static pack_t	out;
static char	msg[MSG_LEN];
static bool	Log = false;
static bool	NoPlayersEnteredYet = true;
static bool	lock = false;
static void	Wait_for_new_players(void);

extern int	login_in_progress;

void Send_meta_server(void);

int main(int argc, char *argv[])
{
    struct hostent *hinfo;
    struct passwd *pwent;

#ifdef apollo
    signal(SIGFPE, SIG_IGN);
#endif
    Argc = argc; Argv = argv;
    init_error(argv[0]);
    srand(time((time_t *)0));		/* Take seed from timer. */
    Parser(argc, argv);
    Make_table();			/* Make trigonometric tables */
    Compute_gravity();
    Find_base_direction();

    /* Allocate memory for players, shots and messages */
    Alloc_players(World.NumBases + MAX_PSEUDO_PLAYERS);
    Alloc_shots(MAX_TOTAL_SHOTS);

    Make_ships();

    if (BIT(World.rules->mode, TEAM_PLAY)) {
	int i;
	for(i=0; i < World.NumTreasures; i++)
	    Make_ball(-1,
		      World.treasures[i].pos.x * BLOCK_SZ+(BLOCK_SZ/2),
		      World.treasures[i].pos.y * BLOCK_SZ+10,
		      false, i);
    }

    /*
     * Get server's official name.
     */
    gethostname(Server.host, 80);
    if ((hinfo = gethostbyname(Server.host)) == NULL) {
	error("gethostbyname %s", Server.host);
    } else {
	strcpy(Server.host, hinfo->h_name);
    }

    /*
     * Get owners login name.
     */
    pwent = getpwuid(geteuid());
    strcpy(Server.name, pwent->pw_name);

    /*
     * Initialize core of out packet.
     */
    out.core.magic = htonl(MAGIC);
    out.core.port = htonl(contactPort);
    strcpy(out.core.realname, Server.name);

    /*
     * Log, if enabled.
     */
    if ((strcmp(Server.name, "kenrsc") == 0) ||
	(strcmp(Server.name, "bjoerns") == 0))
	Log = false;
    Log_game("START");			/* Log start */

    /*
     * Create a socket which we can listen on.
     */
    if ((Socket = CreateDgramSocket(contactPort)) == -1) {
	error("Could not create Dgram socket");
	End_game();
    }
    SetSocketNonBlocking(Socket, 1);

    SetTimeout(0, 0);

    if (Setup_net_server(World.NumBases) == -1) {
	End_game();
    }

    /* 
     * Report to Meta server
     */
   
    Send_meta_server();
    
    /*
     * If the server is not in raw mode it should run only if
     * there are players logged in.
     */
    if (!RawMode) {
	signal(SIGALRM, Handle_signal);	/* Get first client, then proceed. */
	alarm(5*60);			/* Signal me in 5 minutes. */
	Wait_for_new_players();
	alarm(0);
	signal(SIGALRM, SIG_IGN);
	SetTimeout(0, 0);
    }
    signal(SIGHUP, Handle_signal);
    signal(SIGTERM, Handle_signal);
    signal(SIGINT, Handle_signal);
    
    Main_Loop();			    /* Entering main loop. */
    /* NEVER REACHED */
    return (-1);
}

void Send_meta_server(void)
{
    char string[MAX_STR_LEN];
    char status[MAX_STR_LEN];	

    Server_info(status);	

    sprintf(string,
	    "add server %s\n"
	    "add users %d\n"
	    "add version %s\n"
	    "add map %s\n"
	    "add port %d\n"
	    "add status ",
	    Server.host, NumPlayers - NumRobots, 
	    VERSION, World.name, contactPort);
    if (strlen(string) + strlen(status) >= sizeof(string)) {
	/* Prevent string overflow */
	strcpy(&status[sizeof(string) - (strlen(string) + 2)], "\n");
    }
    strcat(string, status);

    DgramSend(Socket, META_HOST, META_PORT, string, strlen(string)+1);
}

/*
 * Main loop.
 */
void Main_Loop(void)
{
    extern void		Loop_delay(void);
    register int	i, x;
    int			main_loops = 0, lastLoops;
    time_t		currentTime, lastPlayerCheckTime = 0;
    time_t		lastMetaCheckTime = 0;
    bool		playerQuit;

#ifndef SILENT
    printf("Server runs at %d frames per second\n", framesPerSecond);
#endif

    setjmp(SavedEnv);

    SetTimeout(0, 0);

    while (NoQuit
	   || NumPlayers - NumPseudoPlayers > NumRobots
	   || NoPlayersEnteredYet) {
	
	currentTime = time(NULL);
	main_loops++;

#define CHECK_FOR_NEW_PLAYERS 5
#define GIVE_META_SERVER_A_HINT 180

	if (currentTime - lastMetaCheckTime >= GIVE_META_SERVER_A_HINT) {
	    lastMetaCheckTime = currentTime;
	    Send_meta_server();
	}

	if (NumPlayers == NumRobots
	    || (login_in_progress && main_loops % 3 == 0)
	    || currentTime - lastPlayerCheckTime >= CHECK_FOR_NEW_PLAYERS) {
	    lastPlayerCheckTime = currentTime;

	    if (NumPlayers == NumRobots && !RawMode) {
		block_timer();
		Wait_for_new_players();
		allow_timer();
	    } else
		Check_new_players();
	}
	
	Update_objects();
	
	if (Shutdown > 0)	/* Check for possible shutdown, the */
	    Shutdown--;		/* server will shutdown when Shutdown */
	else			/* (a counter) reaches 0.  If the */
	    if (Shutdown == 0)	/* counter is < 0 then now shutdown */
		End_game();	/* is in progress. */
	
	if ((main_loops % UPDATES_PR_FRAME) == 0) {
	    Frame_update();
	    Loop_delay();
	}

	Check_client_input();
    }

    End_game();
}


/*
 * Wait for a player to show up.
 */
static void Wait_for_new_players(void)
{
    int			new_players = false;
    time_t		start_time,
			start_loops,
			milli_delta;
    struct timeval	tv;

    gettimeofday(&tv, NULL);
    start_time = tv.tv_sec;
    start_loops = loops;
    while (new_players == false) {
	if (login_in_progress == 0) {
	    SetTimeout(CHECK_FOR_NEW_PLAYERS, 0);
	} else {
	    SetTimeout(0, 100 * 1000);
	}
	if ((new_players = Check_new_players()) == false) {
	    gettimeofday(&tv, NULL);
	    milli_delta = (tv.tv_sec - start_time) * 1000 + tv.tv_usec / 1000;
	    loops = start_loops + (FPS * milli_delta) / 1000;
	}
    }
    SetTimeout(0, 0);
}

/*
 *  Last function, exit with grace.
 */
void End_game(void)
{
    int		i;
    player	*pl;
    char	string[50];

    if (Shutdown == 0) {
	error("Shutting down...");
    }

    while (NumPlayers > 0) {	/* Kick out all remaining players */
	pl = Players[NumPlayers - 1];
	if (pl->conn == NOT_CONNECTED) {
	    Delete_player(NumPlayers - 1);
	} else {
	    Destroy_connection(pl->conn, __FILE__, __LINE__);
	}
    }

    /* Tell meta serve that we are gone */
    sprintf(string,"server %s\nremove",Server.host);
    DgramSend(Socket, META_HOST, META_PORT, string, sizeof(string));

    SocketClose(Socket);
    Free_players();
    Free_ships();
    Free_shots();
    Free_map();
    Log_game("END");			    /* Log end */

    exit (0);
}


void Dump_pack(core_pack_t *p)
{
    printf("\nDUMP OF PACK:\n");
    printf("=============\n");
    printf("TYPE:	%d\n", p->type);
    printf("REALNAME:	%s\n", p->realname);
    printf("PORT:	%ld\n", ntohl(p->port));
    printf("MAGIC:	%lx\n", ntohl(p->magic));
}


bool Check_new_players(void)
{
    int			i,
    			out_size = 0,
    			team,
    			login_port,
    			bytes;
    char		*in_host;
    static bool		lock = false;
    bool		answer = false,
			new_players = false;
    char		*str;
    pack_t		in;
    player		*pl;


    switch (Check_new_connections())
    {
    case -1:
	/* Some connection error */
	break;
    case 0:
	/* Normal stuff */
	break;
    case 1:
	/* New player entered the game */
	NoPlayersEnteredYet = false;
	updateScores = true;
	new_players = true;
	break;
    }

    if (!SocketReadable(Socket))	/* No-one tried to connect. */
	return (new_players);

    /*
     * Someone connected to us, now try and deschiffer the message :)
     */
    errno = 0;
    if ((bytes = DgramReceiveAny(Socket, (char *)&in, sizeof(pack_t))) <= 0) {
	if (errno != EWOULDBLOCK && errno != EINTR) {
	    error("SocketRead (pack from %s)", DgramLastaddr());
	    GetSocketError(Socket);
	}
	return (new_players);
    }

    /*
     * Get hostname.
     */
    in_host = DgramLastaddr();

    /*
     * Determine if we can talk with this hand-shake program.
     */
    if (ntohl(in.core.magic) != MAGIC) {
#ifndef	SILENT
	error("Incompatible packet received from %s", in_host);
#endif
	return (new_players);
    }


    /*
     * Now decode the packet type field and do something witty.
     * (Note, s and r is short for send and reply.)
     */
    out.core.type	= CORE_pack;
    out.core.status	= SUCCESS;
    answer		= true;
    out_size		= sizeof(core_pack_t);

    switch (in.core.type) {

    case ENTER_GAME_pack:	{
	/*
	 * Someone wants to enter the game.
	 */
	reply_pack_t		*s = &out.reply;
	enter_game_pack_t	*r = &in.enter;

	/*
	 * Bad input parameters?
	 */
	r->display[sizeof(r->display) - 1] = '\0'; 
	r->realname[sizeof(r->realname) - 1] = '\0'; 
	r->nick[sizeof(r->nick) - 1] = '\0'; 
	if (r->nick[0] == 0
	    || r->realname[0] == 0
	    || r->display[0] == 0
	    || r->nick[0] < 'A'
	    || r->nick[0] > 'Z') {
	    s->status = E_INVAL;
	    goto switch_end;
	}

	/*
	 * Game locked?
	 */
	if (lock && !Owner(r->realname)) {
	    s->status = E_GAME_LOCKED;
	    goto switch_end;
	}
	
	/*
	 * Is the game full?
	 */
	if (NumPlayers >= World.NumBases) {
	    s->status = E_GAME_FULL;
	    goto switch_end;
	}

	/*
	 * Maybe don't have enough room for player on that team?
	 */
	team = htons(r->team);
	if (BIT(World.rules->mode, TEAM_PLAY)) {
	    if (team == TEAM_NOT_SET || team >= MAX_TEAMS || team < 0) {
		s->status = E_TEAM_NOT_SET;
		goto switch_end;
	    }
	    else if (World.teams[team].NumMembers
		  >= World.teams[team].NumBases) {
		s->status = E_TEAM_FULL;
		goto switch_end;
	    }
	}
	    
	/*
	 * All names must be unique (so we know who we're talking about).
	 */
	for (i=0; i<NumPlayers; i++) {
	    if (strcasecmp(Players[i]->name, r->nick) == 0) {
		s->status = E_IN_USE;
		goto switch_end;
	    }
	}

	/*
	 * Find a port for the client to connect to.
	 */
	if ((login_port = Setup_connection(r->realname,
	    r->nick, r->display, team, in_host)) == -1) {
	    s->status = E_SOCKET;
	    goto switch_end;
	}

	/*
	 * Tell the client which port to use for logging in.
	 */
	sprintf(s->str, "%d", login_port);
	out_size += strlen(s->str) + 1;
    }
	break;


    case REPORT_STATUS_pack:	{
	/*
	 * Someone asked for information.
	 */
	reply_pack_t		*s = &out.reply;
	report_status_pack_t	*r = &in.command;

#ifndef	SILENT
	printf("%s asked for info about current game.\n", r->realname);
#endif
	Server_info(s->str);
	out_size += strlen(s->str);
    }
	break;

	
    case MESSAGE_pack:	{
	/*
	 * Someone wants to transmit a message to the server.
	 */
	reply_pack_t	*s = &out.reply;
	message_pack_t	*r = &in.command;

	if (!Owner(r->realname)) {
	    s->status = E_NOT_OWNER;
	} else {
	    sprintf(msg,
		    "	      <<< MESSAGE FROM ABOVE (%s) >>>	    \"%s\"",
		    r->realname, r->arg_str);
	    Set_message(msg);
	}
    }
	break;


    case LOCK_GAME_pack:	{
	/*
	 * Someone wants to lock the game so that no more players can enter.
	 */
	reply_pack_t		*s = &out.reply;
	lock_game_pack_t	*r = &in.command;

	if (!Owner(r->realname)) {
	    s->status = E_NOT_OWNER;
	} else {
	    lock = !lock;
	}
    }
	break;

	
    case CONTACT_pack:	{
	/*
	 * Got contact message from client.
	 */
	reply_pack_t	*s = &out.reply;
	contact_pack_t	*r = &in.core;

	D(printf("Got CONTACT from %s.\n", in_host));
    }
	break;


    case SHUTDOWN_pack:	{
	/*
	 * Shutdown the entire server.
	 */
	reply_pack_t	*s = &out.reply;
	shutdown_pack_t	*r = &in.command;

	if (!Owner(r->realname)) {
	    s->status = E_NOT_OWNER;
	    goto switch_end;
	}

	Shutdown = ntohl(r->arg_int);

	if (Shutdown == 0) {
	    sprintf(msg, "|*******| SHUTDOWN STOPPED (%s) |*******| \"%s\"",
		    r->realname, r->arg_str);
	    Shutdown = -1;
	} else {
	    sprintf(msg, "|*******| SHUTTING DOWN (%s) |*******| \"%s\"",
		    r->realname, r->arg_str);

	    ShutdownDelay = Shutdown;
	}
	Set_message(msg);
    }
	break;

	
    case KICK_PLAYER_pack:	{
	/*
	 * Kick someone from the game.
	 */
	reply_pack_t		*s = &out.reply;
	kick_player_pack_t	*r = &in.command;
	byte			found = -1;

	if (!Owner(r->realname)) {
	    s->status = E_NOT_OWNER;
	    goto switch_end;
	}

	for (i=0; i<NumPlayers; i++) {
	    if (strcasecmp(r->arg_str, Players[i]->name) == 0
		|| strcasecmp(r->arg_str, Players[i]->realname) == 0)
		found = i;
	}
	if (found == -1) {
	    s->status = E_NOT_FOUND;
	} else {
	    sprintf(msg, "\"%s\" upset the gods and was kicked out "
		    "of the game.", Players[found]->name);
	    Set_message(msg);
	    if (Players[found]->conn == NOT_CONNECTED) {
		Delete_player(found);
	    } else {
		Destroy_connection(Players[found]->conn, __FILE__, __LINE__);
	    }
	    updateScores = true;
	    goto switch_end;
	}
    }
	break;

	
    default:
	/*
	 * Incorrect packet type.
	 */
	answer	= false;

	error("Incorrect packet type from %s.\n"
	      "(Probably occured due to incompatibility between handshake\n"
	      "version and server version, someone should upgrade)", in_host);
    }

 switch_end:
    if (answer && (DgramSend(Socket, in_host, ntohl(in.core.port),
			     (char *)&out, out_size)) == -1) {
	error("Could not send request to client at %s.", in_host);
    }

    return new_players;
}

/*
 * Return status for server 
*/
void Server_info(char *str)
{
    int i, j, k;
    player *pl, **order, *best = NULL;
    float ratio, best_ratio;
    char name[MAX_CHARS];
    char lblstr[MAX_CHARS];

    sprintf(str,
	    "SERVER VERSION...: %s\n"
	    "STARTED BY.......: %s\n"
	    "STATUS...........: %s\n"
	    "MAX SPEED........: %d fps\n"
	    "WORLD (%3dx%3d)..: %s\n"
	    "      AUTHOR.....: %s\n"
	    "PLAYERS (%2d/%2d)..:\n",
	    TITLE,
	    Server.name,
	    lock && Shutdown == -1 ? "locked" :
	    !lock && Shutdown != -1 ? "shutting down" :
	    lock && Shutdown != -1 ? "locked and shutting down" : "ok",
	    FPS,
	    World.x, World.y, World.name, World.author,
	    NumPlayers, World.NumBases);

    if (NumPlayers > 0)
	strcat(str,
	       "\nNO:  TM: NAME:             LIFE:   SC:    PLAYER:\n"
	       "-------------------------------------------------\n");
	
    if ((order = (player **) malloc(NumPlayers * sizeof(player *))) == NULL) {
	error("No memory for order");
	return;
    }
    for (i=0; i<NumPlayers; i++) {
	pl = Players[i];
	ratio = (float) pl->score / (pl->life + 1);
	if (best == NULL
	    || ratio > best_ratio) {
	    best_ratio = ratio;
	    best = pl;
	}
	for (j = 0; j < i; j++) {
	    if (order[j]->score < pl->score) {
		for (k = i; k > j; k--) {
		    order[k] = order[k - 1];
		}
		break;
	    } 
	}
	order[j] = pl;
    }
    for (i=0; i<NumPlayers; i++) {
	pl = order[i];
	strcpy(name, pl->name);
	if (pl->robot_mode != RM_NOT_ROBOT
	    && pl->robot_lock == LOCK_PLAYER) {
	    sprintf(name + strlen(name), " (%s)",
		Players[GetInd[pl->robot_lock_id]]->name);
	    if (strlen(name) >= 19) {
		strcpy(&name[17], ")");
	    }
	}
	sprintf(lblstr, "%c%c %-19s%03d%6d",
		(pl == best) ? '*' : pl->mychar,
		(pl->team == TEAM_NOT_SET) ? ' ' : pl->team+'0',
		name, pl->life, pl->score);
	sprintf(msg, "%2d... %-36s%s@%s\n",
		i+1, lblstr, pl->realname,
		pl->robot_mode == RM_NOT_ROBOT
		? pl->dispname
		: "robots.org");
	if (strlen(msg) + strlen(str) >= MAX_STR_LEN)
	    break;
	strcat(str, msg);
    }
    free(order);
}

/*
 * Returns true if <name> has owner status of this server.
 */
bool Owner(char *name)
{
    if ((strcmp(name, Server.name)  == 0)
	|| (strcmp(name, "kenrsc")  == 0)
	|| (strcmp(name, "bjoerns") == 0)
	|| (strcmp(name, "root")    == 0))
	return (true);
    else
	return (false);
}


void Handle_signal(int sig_no)
{
    /* Tell meta serve that we are gone */
    char	string[50];
    sprintf(string,"server %s\nremove",Server.host);
    DgramSend(Socket, META_HOST, META_PORT, string, sizeof(string));
    
    switch (sig_no) {
    case SIGALRM:
	error("First player has yet to show his butt, I'm bored... Bye!");
	SocketClose(Socket);
	break;

    case SIGTERM:
	error("Caught SIGTERM, terminating.");
	End_game();
	break;
    case SIGHUP:
	error("Caught SIGHUP, terminating.");
	End_game();
	break;
    case SIGINT:
	error("Caught SIGINT, terminating.");
	End_game();
	break;

    default:
	error("Caught unkown signal: %d", sig_no);
	break;
    }

    exit(sig_no);
}


void Log_game(char *heading)
{
#ifdef LOG
    char str[1024];
    FILE *fp;

    if (!Log)
	return;

    sprintf(str,
	    "%s (%s) - %s@%s, map '%s' (%d)\n",
	    heading, VERSION, Server.name, Server.host,
	    World.name, Server.max_num);

    if ((fp=fopen(LOGFILE, "a")) == NULL) {	/* Couldn't open file */
	error("Couldn't open log file, contact " LOCALGURU "");
	return;
    }

    fputs(str, fp);
    fclose(fp);
#endif
}
